# PytgVoIP - Telegram VoIP Library for Python
# Copyright (C) 2019 bakatrouble <https://github.com/bakatrouble>
#
# This file is part of PytgVoIP.
#
# PytgVoIP is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PytgVoIP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with PytgVoIP.  If not, see <http://www.gnu.org/licenses/>.


import hashlib
import time
from typing import List, Union

"""
.. module:: utils
    :synopsis: Utility functions for pytgvoip
"""

twoe1984 = 1 << 1984  # 2^1984 https://core.telegram.org/api/end-to-end#sending-a-request
emojis = (
    ('\U0001f609', 'WINKING_FACE'),
    ('\U0001f60d', 'SMILING_FACE_WITH_HEART_EYES'),
    ('\U0001f61b', 'FACE_WITH_TONGUE'),
    ('\U0001f62d', 'LOUDLY_CRYING_FACE'),
    ('\U0001f631', 'FACE_SCREAMING_IN_FEAR'),
    ('\U0001f621', 'POUTING_FACE'),
    ('\U0001f60e', 'SMILING_FACE_WITH_SUNGLASSES'),
    ('\U0001f634', 'SLEEPING_FACE'),
    ('\U0001f635', 'DIZZY_FACE'),
    ('\U0001f608', 'SMILING_FACE_WITH_HORNS'),
    ('\U0001f62c', 'GRIMACING_FACE'),
    ('\U0001f607', 'SMILING_FACE_WITH_HALO'),
    ('\U0001f60f', 'SMIRKING_FACE'),
    ('\U0001f46e', 'POLICE_OFFICER'),
    ('\U0001f477', 'CONSTRUCTION_WORKER'),
    ('\U0001f482', 'GUARD'),
    ('\U0001f476', 'BABY'),
    ('\U0001f468', 'MAN'),
    ('\U0001f469', 'WOMAN'),
    ('\U0001f474', 'OLD_MAN'),
    ('\U0001f475', 'OLD_WOMAN'),
    ('\U0001f63b', 'SMILING_CAT_FACE_WITH_HEART_EYES'),
    ('\U0001f63d', 'KISSING_CAT_FACE'),
    ('\U0001f640', 'WEARY_CAT_FACE'),
    ('\U0001f47a', 'GOBLIN'),
    ('\U0001f648', 'SEE_NO_EVIL_MONKEY'),
    ('\U0001f649', 'HEAR_NO_EVIL_MONKEY'),
    ('\U0001f64a', 'SPEAK_NO_EVIL_MONKEY'),
    ('\U0001f480', 'SKULL'),
    ('\U0001f47d', 'ALIEN'),
    ('\U0001f4a9', 'PILE_OF_POO'),
    ('\U0001f525', 'FIRE'),
    ('\U0001f4a5', 'COLLISION'),
    ('\U0001f4a4', 'ZZZ'),
    ('\U0001f442', 'EAR'),
    ('\U0001f440', 'EYES'),
    ('\U0001f443', 'NOSE'),
    ('\U0001f445', 'TONGUE'),
    ('\U0001f444', 'MOUTH'),
    ('\U0001f44d', 'THUMBS_UP'),
    ('\U0001f44e', 'THUMBS_DOWN'),
    ('\U0001f44c', 'OK_HAND'),
    ('\U0001f44a', 'ONCOMING_FIST'),
    ('\u270c', 'VICTORY_HAND'),
    ('\u270b', 'RAISED_HAND'),
    ('\U0001f450', 'OPEN_HANDS'),
    ('\U0001f446', 'BACKHAND_INDEX_POINTING_UP'),
    ('\U0001f447', 'BACKHAND_INDEX_POINTING_DOWN'),
    ('\U0001f449', 'BACKHAND_INDEX_POINTING_RIGHT'),
    ('\U0001f448', 'BACKHAND_INDEX_POINTING_LEFT'),
    ('\U0001f64f', 'FOLDED_HANDS'),
    ('\U0001f44f', 'CLAPPING_HANDS'),
    ('\U0001f4aa', 'FLEXED_BICEPS'),
    ('\U0001f6b6', 'PERSON_WALKING'),
    ('\U0001f3c3', 'PERSON_RUNNING'),
    ('\U0001f483', 'WOMAN_DANCING'),
    ('\U0001f46b', 'MAN_AND_WOMAN_HOLDING_HANDS'),
    ('\U0001f46a', 'FAMILY'),
    ('\U0001f46c', 'TWO_MEN_HOLDING_HANDS'),
    ('\U0001f46d', 'TWO_WOMEN_HOLDING_HANDS'),
    ('\U0001f485', 'NAIL_POLISH'),
    ('\U0001f3a9', 'TOP_HAT'),
    ('\U0001f451', 'CROWN'),
    ('\U0001f452', 'WOMAN_S_HAT'),
    ('\U0001f45f', 'RUNNING_SHOE'),
    ('\U0001f45e', 'MAN_S_SHOE'),
    ('\U0001f460', 'HIGH_HEELED_SHOE'),
    ('\U0001f455', 'T_SHIRT'),
    ('\U0001f457', 'DRESS'),
    ('\U0001f456', 'JEANS'),
    ('\U0001f459', 'BIKINI'),
    ('\U0001f45c', 'HANDBAG'),
    ('\U0001f453', 'GLASSES'),
    ('\U0001f380', 'RIBBON'),
    ('\U0001f484', 'LIPSTICK'),
    ('\U0001f49b', 'YELLOW_HEART'),
    ('\U0001f499', 'BLUE_HEART'),
    ('\U0001f49c', 'PURPLE_HEART'),
    ('\U0001f49a', 'GREEN_HEART'),
    ('\U0001f48d', 'RING'),
    ('\U0001f48e', 'GEM_STONE'),
    ('\U0001f436', 'DOG_FACE'),
    ('\U0001f43a', 'WOLF_FACE'),
    ('\U0001f431', 'CAT_FACE'),
    ('\U0001f42d', 'MOUSE_FACE'),
    ('\U0001f439', 'HAMSTER_FACE'),
    ('\U0001f430', 'RABBIT_FACE'),
    ('\U0001f438', 'FROG_FACE'),
    ('\U0001f42f', 'TIGER_FACE'),
    ('\U0001f428', 'KOALA'),
    ('\U0001f43b', 'BEAR_FACE'),
    ('\U0001f437', 'PIG_FACE'),
    ('\U0001f42e', 'COW_FACE'),
    ('\U0001f417', 'BOAR'),
    ('\U0001f434', 'HORSE_FACE'),
    ('\U0001f411', 'EWE'),
    ('\U0001f418', 'ELEPHANT'),
    ('\U0001f43c', 'PANDA_FACE'),
    ('\U0001f427', 'PENGUIN'),
    ('\U0001f425', 'FRONT_FACING_BABY_CHICK'),
    ('\U0001f414', 'CHICKEN'),
    ('\U0001f40d', 'SNAKE'),
    ('\U0001f422', 'TURTLE'),
    ('\U0001f41b', 'BUG'),
    ('\U0001f41d', 'HONEYBEE'),
    ('\U0001f41c', 'ANT'),
    ('\U0001f41e', 'LADY_BEETLE'),
    ('\U0001f40c', 'SNAIL'),
    ('\U0001f419', 'OCTOPUS'),
    ('\U0001f41a', 'SPIRAL_SHELL'),
    ('\U0001f41f', 'FISH'),
    ('\U0001f42c', 'DOLPHIN'),
    ('\U0001f40b', 'WHALE'),
    ('\U0001f410', 'GOAT'),
    ('\U0001f40a', 'CROCODILE'),
    ('\U0001f42b', 'TWO_HUMP_CAMEL'),
    ('\U0001f340', 'FOUR_LEAF_CLOVER'),
    ('\U0001f339', 'ROSE'),
    ('\U0001f33b', 'SUNFLOWER'),
    ('\U0001f341', 'MAPLE_LEAF'),
    ('\U0001f33e', 'SHEAF_OF_RICE'),
    ('\U0001f344', 'MUSHROOM'),
    ('\U0001f335', 'CACTUS'),
    ('\U0001f334', 'PALM_TREE'),
    ('\U0001f333', 'DECIDUOUS_TREE'),
    ('\U0001f31e', 'SUN_WITH_FACE'),
    ('\U0001f31a', 'NEW_MOON_FACE'),
    ('\U0001f319', 'CRESCENT_MOON'),
    ('\U0001f30e', 'GLOBE_SHOWING_AMERICAS'),
    ('\U0001f30b', 'VOLCANO'),
    ('\u26a1', 'HIGH_VOLTAGE'),
    ('\u2614', 'UMBRELLA_WITH_RAIN_DROPS'),
    ('\u2744', 'SNOWFLAKE'),
    ('\u26c4', 'SNOWMAN_WITHOUT_SNOW'),
    ('\U0001f300', 'CYCLONE'),
    ('\U0001f308', 'RAINBOW'),
    ('\U0001f30a', 'WATER_WAVE'),
    ('\U0001f393', 'GRADUATION_CAP'),
    ('\U0001f386', 'FIREWORKS'),
    ('\U0001f383', 'JACK_O_LANTERN'),
    ('\U0001f47b', 'GHOST'),
    ('\U0001f385', 'SANTA_CLAUS'),
    ('\U0001f384', 'CHRISTMAS_TREE'),
    ('\U0001f381', 'WRAPPED_GIFT'),
    ('\U0001f388', 'BALLOON'),
    ('\U0001f52e', 'CRYSTAL_BALL'),
    ('\U0001f3a5', 'MOVIE_CAMERA'),
    ('\U0001f4f7', 'CAMERA'),
    ('\U0001f4bf', 'OPTICAL_DISK'),
    ('\U0001f4bb', 'LAPTOP_COMPUTER'),
    ('\u260e', 'TELEPHONE'),
    ('\U0001f4e1', 'SATELLITE_ANTENNA'),
    ('\U0001f4fa', 'TELEVISION'),
    ('\U0001f4fb', 'RADIO'),
    ('\U0001f509', 'SPEAKER_MEDIUM_VOLUME'),
    ('\U0001f514', 'BELL'),
    ('\u23f3', 'HOURGLASS_NOT_DONE'),
    ('\u23f0', 'ALARM_CLOCK'),
    ('\u231a', 'WATCH'),
    ('\U0001f512', 'LOCKED'),
    ('\U0001f511', 'KEY'),
    ('\U0001f50e', 'MAGNIFYING_GLASS_TILTED_RIGHT'),
    ('\U0001f4a1', 'LIGHT_BULB'),
    ('\U0001f526', 'FLASHLIGHT'),
    ('\U0001f50c', 'ELECTRIC_PLUG'),
    ('\U0001f50b', 'BATTERY'),
    ('\U0001f6bf', 'SHOWER'),
    ('\U0001f6bd', 'TOILET'),
    ('\U0001f527', 'WRENCH'),
    ('\U0001f528', 'HAMMER'),
    ('\U0001f6aa', 'DOOR'),
    ('\U0001f6ac', 'CIGARETTE'),
    ('\U0001f4a3', 'BOMB'),
    ('\U0001f52b', 'PISTOL'),
    ('\U0001f52a', 'KITCHEN_KNIFE'),
    ('\U0001f48a', 'PILL'),
    ('\U0001f489', 'SYRINGE'),
    ('\U0001f4b0', 'MONEY_BAG'),
    ('\U0001f4b5', 'DOLLAR_BANKNOTE'),
    ('\U0001f4b3', 'CREDIT_CARD'),
    ('\u2709', 'ENVELOPE'),
    ('\U0001f4eb', 'CLOSED_MAILBOX_WITH_RAISED_FLAG'),
    ('\U0001f4e6', 'PACKAGE'),
    ('\U0001f4c5', 'CALENDAR'),
    ('\U0001f4c1', 'FILE_FOLDER'),
    ('\u2702', 'SCISSORS'),
    ('\U0001f4cc', 'PUSHPIN'),
    ('\U0001f4ce', 'PAPERCLIP'),
    ('\u2712', 'BLACK_NIB'),
    ('\u270f', 'PENCIL'),
    ('\U0001f4d0', 'TRIANGULAR_RULER'),
    ('\U0001f4da', 'BOOKS'),
    ('\U0001f52c', 'MICROSCOPE'),
    ('\U0001f52d', 'TELESCOPE'),
    ('\U0001f3a8', 'ARTIST_PALETTE'),
    ('\U0001f3ac', 'CLAPPER_BOARD'),
    ('\U0001f3a4', 'MICROPHONE'),
    ('\U0001f3a7', 'HEADPHONE'),
    ('\U0001f3b5', 'MUSICAL_NOTE'),
    ('\U0001f3b9', 'MUSICAL_KEYBOARD'),
    ('\U0001f3bb', 'VIOLIN'),
    ('\U0001f3ba', 'TRUMPET'),
    ('\U0001f3b8', 'GUITAR'),
    ('\U0001f47e', 'ALIEN_MONSTER'),
    ('\U0001f3ae', 'VIDEO_GAME'),
    ('\U0001f0cf', 'JOKER'),
    ('\U0001f3b2', 'GAME_DIE'),
    ('\U0001f3af', 'DIRECT_HIT'),
    ('\U0001f3c8', 'AMERICAN_FOOTBALL'),
    ('\U0001f3c0', 'BASKETBALL'),
    ('\u26bd', 'SOCCER_BALL'),
    ('\u26be', 'BASEBALL'),
    ('\U0001f3be', 'TENNIS'),
    ('\U0001f3b1', 'POOL_8_BALL'),
    ('\U0001f3c9', 'RUGBY_FOOTBALL'),
    ('\U0001f3b3', 'BOWLING'),
    ('\U0001f3c1', 'CHEQUERED_FLAG'),
    ('\U0001f3c7', 'HORSE_RACING'),
    ('\U0001f3c6', 'TROPHY'),
    ('\U0001f3ca', 'PERSON_SWIMMING'),
    ('\U0001f3c4', 'PERSON_SURFING'),
    ('\u2615', 'HOT_BEVERAGE'),
    ('\U0001f37c', 'BABY_BOTTLE'),
    ('\U0001f37a', 'BEER_MUG'),
    ('\U0001f377', 'WINE_GLASS'),
    ('\U0001f374', 'FORK_AND_KNIFE'),
    ('\U0001f355', 'PIZZA'),
    ('\U0001f354', 'HAMBURGER'),
    ('\U0001f35f', 'FRENCH_FRIES'),
    ('\U0001f357', 'POULTRY_LEG'),
    ('\U0001f371', 'BENTO_BOX'),
    ('\U0001f35a', 'COOKED_RICE'),
    ('\U0001f35c', 'STEAMING_BOWL'),
    ('\U0001f361', 'DANGO'),
    ('\U0001f373', 'COOKING'),
    ('\U0001f35e', 'BREAD'),
    ('\U0001f369', 'DOUGHNUT'),
    ('\U0001f366', 'SOFT_ICE_CREAM'),
    ('\U0001f382', 'BIRTHDAY_CAKE'),
    ('\U0001f370', 'SHORTCAKE'),
    ('\U0001f36a', 'COOKIE'),
    ('\U0001f36b', 'CHOCOLATE_BAR'),
    ('\U0001f36d', 'LOLLIPOP'),
    ('\U0001f36f', 'HONEY_POT'),
    ('\U0001f34e', 'RED_APPLE'),
    ('\U0001f34f', 'GREEN_APPLE'),
    ('\U0001f34a', 'TANGERINE'),
    ('\U0001f34b', 'LEMON'),
    ('\U0001f352', 'CHERRIES'),
    ('\U0001f347', 'GRAPES'),
    ('\U0001f349', 'WATERMELON'),
    ('\U0001f353', 'STRAWBERRY'),
    ('\U0001f351', 'PEACH'),
    ('\U0001f34c', 'BANANA'),
    ('\U0001f350', 'PEAR'),
    ('\U0001f34d', 'PINEAPPLE'),
    ('\U0001f346', 'EGGPLANT'),
    ('\U0001f345', 'TOMATO'),
    ('\U0001f33d', 'EAR_OF_CORN'),
    ('\U0001f3e1', 'HOUSE_WITH_GARDEN'),
    ('\U0001f3e5', 'HOSPITAL'),
    ('\U0001f3e6', 'BANK'),
    ('\u26ea', 'CHURCH'),
    ('\U0001f3f0', 'CASTLE'),
    ('\u26fa', 'TENT'),
    ('\U0001f3ed', 'FACTORY'),
    ('\U0001f5fb', 'MOUNT_FUJI'),
    ('\U0001f5fd', 'STATUE_OF_LIBERTY'),
    ('\U0001f3a0', 'CAROUSEL_HORSE'),
    ('\U0001f3a1', 'FERRIS_WHEEL'),
    ('\u26f2', 'FOUNTAIN'),
    ('\U0001f3a2', 'ROLLER_COASTER'),
    ('\U0001f6a2', 'SHIP'),
    ('\U0001f6a4', 'SPEEDBOAT'),
    ('\u2693', 'ANCHOR'),
    ('\U0001f680', 'ROCKET'),
    ('\u2708', 'AIRPLANE'),
    ('\U0001f681', 'HELICOPTER'),
    ('\U0001f682', 'LOCOMOTIVE'),
    ('\U0001f68b', 'TRAM_CAR'),
    ('\U0001f68e', 'TROLLEYBUS'),
    ('\U0001f68c', 'BUS'),
    ('\U0001f699', 'SPORT_UTILITY_VEHICLE'),
    ('\U0001f697', 'AUTOMOBILE'),
    ('\U0001f695', 'TAXI'),
    ('\U0001f69b', 'ARTICULATED_LORRY'),
    ('\U0001f6a8', 'POLICE_CAR_LIGHT'),
    ('\U0001f694', 'ONCOMING_POLICE_CAR'),
    ('\U0001f692', 'FIRE_ENGINE'),
    ('\U0001f691', 'AMBULANCE'),
    ('\U0001f6b2', 'BICYCLE'),
    ('\U0001f6a0', 'MOUNTAIN_CABLEWAY'),
    ('\U0001f69c', 'TRACTOR'),
    ('\U0001f6a6', 'VERTICAL_TRAFFIC_LIGHT'),
    ('\u26a0', 'WARNING'),
    ('\U0001f6a7', 'CONSTRUCTION'),
    ('\u26fd', 'FUEL_PUMP'),
    ('\U0001f3b0', 'SLOT_MACHINE'),
    ('\U0001f5ff', 'MOAI'),
    ('\U0001f3aa', 'CIRCUS_TENT'),
    ('\U0001f3ad', 'PERFORMING_ARTS'),
    ('\U0001f1ef\\U0001f1f5', 'JAPAN'),
    ('\U0001f1f0\\U0001f1f7', 'SOUTH_KOREA'),
    ('\U0001f1e9\\U0001f1ea', 'GERMANY'),
    ('\U0001f1e8\\U0001f1f3', 'CHINA'),
    ('\U0001f1fa\\U0001f1f8', 'UNITED_STATES'),
    ('\U0001f1eb\\U0001f1f7', 'FRANCE'),
    ('\U0001f1ea\\U0001f1f8', 'SPAIN'),
    ('\U0001f1ee\\U0001f1f9', 'ITALY'),
    ('\U0001f1f7\\U0001f1fa', 'RUSSIA'),
    ('\U0001f1ec\\U0001f1e7', 'UNITED_KINGDOM'),
    ('1\u20e3', 'KEYCAP_ONE'),
    ('2\u20e3', 'KEYCAP_TWO'),
    ('3\u20e3', 'KEYCAP_THREE'),
    ('4\u20e3', 'KEYCAP_FOUR'),
    ('5\u20e3', 'KEYCAP_FIVE'),
    ('6\u20e3', 'KEYCAP_SIX'),
    ('7\u20e3', 'KEYCAP_SEVEN'),
    ('8\u20e3', 'KEYCAP_EIGHT'),
    ('9\u20e3', 'KEYCAP_NINE'),
    ('0\u20e3', 'KEYCAP_ZERO'),
    ('\U0001f51f', 'KEYCAP_10'),
    ('\u2757', 'EXCLAMATION_MARK'),
    ('\u2753', 'QUESTION_MARK'),
    ('\u2665', 'HEART_SUIT'),
    ('\u2666', 'DIAMOND_SUIT'),
    ('\U0001f4af', 'HUNDRED_POINTS'),
    ('\U0001f517', 'LINK'),
    ('\U0001f531', 'TRIDENT_EMBLEM'),
    ('\U0001f534', 'RED_CIRCLE'),
    ('\U0001f535', 'BLUE_CIRCLE'),
    ('\U0001f536', 'LARGE_ORANGE_DIAMOND'),
    ('\U0001f537', 'LARGE_BLUE_DIAMOND'),
)
common_prime = b'\xC7\x1C\xAE\xB9\xC6\xB1\xC9\x04\x8E\x6C\x52\x2F\x70\xF1\x3F\x73\x98\x0D\x40\x23\x8E\x3E\x21\xC1\x49' \
               b'\x34\xD0\x37\x56\x3D\x93\x0F\x48\x19\x8A\x0A\xA7\xC1\x40\x58\x22\x94\x93\xD2\x25\x30\xF4\xDB\xFA\x33' \
               b'\x6F\x6E\x0A\xC9\x25\x13\x95\x43\xAE\xD4\x4C\xCE\x7C\x37\x20\xFD\x51\xF6\x94\x58\x70\x5A\xC6\x8C\xD4' \
               b'\xFE\x6B\x6B\x13\xAB\xDC\x97\x46\x51\x29\x69\x32\x84\x54\xF1\x8F\xAF\x8C\x59\x5F\x64\x24\x77\xFE\x96' \
               b'\xBB\x2A\x94\x1D\x5B\xCD\x1D\x4A\xC8\xCC\x49\x88\x07\x08\xFA\x9B\x37\x8E\x3C\x4F\x3A\x90\x60\xBE\xE6' \
               b'\x7C\xF9\xA4\xA4\xA6\x95\x81\x10\x51\x90\x7E\x16\x27\x53\xB5\x6B\x0F\x6B\x41\x0D\xBA\x74\xD8\xA8\x4B' \
               b'\x2A\x14\xB3\x14\x4E\x0E\xF1\x28\x47\x54\xFD\x17\xED\x95\x0D\x59\x65\xB4\xB9\xDD\x46\x58\x2D\xB1\x17' \
               b'\x8D\x16\x9C\x6B\xC4\x65\xB0\xD6\xFF\x9C\xA3\x92\x8F\xEF\x5B\x9A\xE4\xE4\x18\xFC\x15\xE8\x3E\xBE\xA0' \
               b'\xF8\x7F\xA9\xFF\x5E\xED\x70\x05\x0D\xED\x28\x49\xF4\x7B\xF9\x59\xD9\x56\x85\x0C\xE9\x29\x85\x1F\x0D' \
               b'\x81\x15\xF6\x35\xB1\x05\xEE\x2E\x4E\x15\xD0\x4B\x24\x54\xBF\x6F\x4F\xAD\xF0\x34\xB1\x04\x03\x11\x9C' \
               b'\xD8\xE3\xB9\x2F\xCC\x5B'


def i2b(value: int) -> bytes:
    """
    Convert integer value to bytes

    Args:
        value (``int``):
            Value to convert

    Returns:
        Resulting ``bytes`` object
    """
    return int.to_bytes(
        value,
        length=(value.bit_length() + 8 - 1) // 8,  # 8 bits per byte,
        byteorder='big',
        signed=False
    )


def b2i(value: bytes) -> int:
    """
    Convert bytes value to integer

    Args:
        value (``bytes``):
            Value to convert

    Returns:
        Resulting ``int`` object
    """
    return int.from_bytes(value, 'big')


def check_dhc(g: int, p: int) -> None:
    """
    Security checks for Diffie-Hellman prime and generator. Ported from Java implementation for Android

    Args:
        g (``int``): DH generator
        p (``int``): DH prime

    Raises:
        :class:`ValueError` if checks are not passed
    """
    if not 2 <= g <= 7:
        raise ValueError()

    if p.bit_length() != 2048 or p < 0:
        raise ValueError()

    if (
            g == 2 and p % 8 != 7 or  # p % 8 = 7 for g = 2
            g == 3 and p % 3 != 2 or  # p % 3 = 2 for g = 3
            g == 5 and p % 5 not in (1, 4) or  # p % 5 = 1 or 4 for g = 5
            g == 6 and p % 24 not in (19, 23) or  # p % 24 = 19 or 23 for g = 6
            g == 7 and p % 7 not in (3, 5, 6)  # p % 7 = 3, 5 or 6 for g = 7
    ):
        raise ValueError()

    if i2b(p) == common_prime:
        return

    # let's assume that (p - 1) / 2 is prime because checking its primality is an expensive operation...


def check_g(g_x: int, p: int) -> None:
    """
    Check g\_ numbers

    Args:
        g_x: g\_ number to check
        p: DH prime

    Raises:
        :class:`ValueError` if checks are not passed
    """
    if not (1 < g_x < p - 1):
        raise ValueError('g_x is invalid (1 < g_x < p - 1 is false)')
    if not (twoe1984 < g_x < p - twoe1984):
        raise ValueError('g_x is invalid (2^1984 < g_x < p - 2^1984 is false)')


def calc_fingerprint(key: bytes) -> int:
    """
    Calculate key fingerprint

    Args:
        key (``bytes``):
            Key to generate fingerprint for

    Returns:
        :class:`int` object representing a key fingerprint
    """
    return int.from_bytes(
        bytes(hashlib.sha1(key).digest()[-8:]), 'little', signed=True
    )


def generate_visualization(key: Union[bytes, int], part2: Union[bytes, int]) -> (List[str], List[str]):
    """
    Generate emoji visualization of key

    https://core.telegram.org/api/end-to-end/voice-calls#key-verification

    Args:
        key (``bytes`` | ``int``):
            Call auth key

        part2 (``bytes`` | ``int``):
            `g_a` value of the caller

    Returns:
        A tuple containing two lists (of emoji strings and of their text representations)
    """
    if isinstance(key, int):
        key = i2b(key)
    if isinstance(part2, int):
        part2 = i2b(part2)

    visualization = []
    visualization_text = []
    vis_src = hashlib.sha256(key + part2).digest()
    for i in range(0, len(vis_src), 8):
        number = vis_src[i:i+8]
        number = i2b(number[0] & 0x7f) + number[1:]
        idx = int.from_bytes(number, 'big') % len(emojis)
        visualization.append(emojis[idx][0])
        visualization_text.append(emojis[idx][1])
    return visualization, visualization_text


def get_real_elapsed_time() -> float:
    """
    Get current performance counter value

    Returns:
        Time to use for measuring call duration
    """
    return time.perf_counter()



# PytgVoIP - Telegram VoIP Library for Python
# Copyright (C) 2019 bakatrouble <https://github.com/bakatrouble>
#
# This file is part of PytgVoIP.
#
# PytgVoIP is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PytgVoIP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with PytgVoIP.  If not, see <http://www.gnu.org/licenses/>.


from enum import Enum
from typing import Optional, List


class NetType(Enum):
    UNKNOWN = ...
    GPRS = ...
    EDGE = ...
    NET_3G = ...
    HSPA = ...
    LTE = ...
    WIFI = ...
    ETHERNET = ...
    OTHER_HIGH_SPEED = ...
    OTHER_LOW_SPEED = ...
    DIALUP = ...
    OTHER_MOBILE = ...


class CallState(Enum):
    WAIT_INIT = ...
    WAIT_INIT_ACK = ...
    ESTABLISHED = ...
    FAILED = ...
    RECONNECTING = ...


class DataSaving(Enum):
    NEVER = ...
    MOBILE = ...
    ALWAYS = ...


class CallError(Enum):
    UNKNOWN = ...
    INCOMPATIBLE = ...
    TIMEOUT = ...
    AUDIO_IO = ...
    PROXY = ...


class Stats:
    bytes_sent_wifi = ...
    bytes_sent_mobile = ...
    bytes_recvd_wifi = ...
    bytes_recvd_mobile = ...


# class AudioInputDevice:
#     _id = ...
#     display_name = ...


# class AudioOutputDevice:
#     _id = ...
#     display_name = ...


class Endpoint:
    _id: int = ...
    ip: str = ...
    ipv6: str = ...
    port: int = ...
    peer_tag: bytes = ...

    def __init__(self, _id: int, ip: str, ipv6: str, port: int, peer_tag: bytes): ...


class VoIPController:
    LIBTGVOIP_VERSION: str = ...
    CONNECTION_MAX_LAYER: int = ...

    def __init__(self, persistent_state_file: str = ''): ...
    def _init(self) -> None: ...
    def start(self) -> None: ...
    def connect(self) -> None: ...
    def set_proxy(self, address: str, port: int, username: str, password: str) -> None: ...
    def set_encryption_key(self, key: bytes, is_outgoing: bool) -> None: ...
    def set_remote_endpoints(self, endpoints: list, allow_p2p: bool, tcp: bool, connection_max_layer: int) -> None: ...
    def get_debug_string(self) -> str: ...
    def set_network_type(self, _type: NetType) -> None: ...
    def set_mic_mute(self, mute: bool) -> None: ...
    def set_config(self, recv_timeout: float, init_timeout: float, data_saving_mode: DataSaving, enable_aec: bool,
                   enable_ns: bool, enable_agc: bool, log_file_path: str, status_dump_path: str,
                   log_packet_stats: bool) -> None: ...
    def debug_ctl(self, request: int, param: int) -> None: ...
    def get_preferred_relay_id(self) -> int: ...
    def get_last_error(self) -> CallError: ...
    def get_stats(self) -> Stats: ...
    def get_debug_log(self) -> str: ...
    def set_audio_output_gain_control_enabled(self, enabled: bool) -> None: ...
    def set_echo_cancellation_strength(self, strength: int) -> None: ...
    def get_peer_capabilities(self) -> int: ...
    def need_rate(self) -> bool: ...

    # @staticmethod
    # def enumerate_audio_inputs() -> List[AudioInputDevice]: ...
    #
    # @staticmethod
    # def enumerate_audio_outputs() -> List[AudioOutputDevice]: ...
    #
    # def set_current_audio_input(self, _id: str) -> None: ...
    # def set_current_audio_output(self, _id: str) -> None: ...
    # def get_current_audio_input_id(self) -> str: ...
    # def get_current_audio_output_id(self) -> str: ...

    def _handle_state_change(self, state: CallState) -> None:
        raise NotImplementedError()

    def _handle_signal_bars_change(self, count: int) -> None:
        raise NotImplementedError()

    def _send_audio_frame_impl(self, length: int) -> bytes: ...
    def _recv_audio_frame_impl(self, frame: bytes) -> None: ...


class VoIPServerConfig:
    @staticmethod
    def set_config(json_string: str): ...


__version__: str = ...



'''
This file is part of Telegram Desktop,
the official desktop application for the Telegram messaging service.

For license and copyright information please follow this link:
https://github.com/telegramdesktop/tdesktop/blob/master/LEGAL
'''
import sys, os, shutil, subprocess

def run(project, arguments, buildType=''):
    scriptPath = os.path.dirname(os.path.realpath(__file__))
    basePath = scriptPath + '/../out/' + buildType

    cmake = ['cmake']
    for arg in arguments:
        if arg == 'debug':
            cmake.append('-DCMAKE_BUILD_TYPE=Debug')
        elif arg != 'force':
            cmake.append(arg)

    if sys.platform == 'win32':
        cmake.append('-AWin32')
    elif sys.platform == 'darwin':
        cmake.append('-GXcode')
    elif buildType:
        cmake.append('-DCMAKE_BUILD_TYPE=' + buildType)
    elif not '-DCMAKE_BUILD_TYPE=Debug' in cmake:
        cmake.append('-DCMAKE_BUILD_TYPE=Release')

    specialTarget = ''
    specialTargetFile = scriptPath + '/../' + project + '/build/target'
    if os.path.isfile(specialTargetFile):
        with open(specialTargetFile, 'r') as f:
            for line in f:
                target = line.strip()
                if len(target) > 0:
                    cmake.append('-DDESKTOP_APP_SPECIAL_TARGET=' + target)

    cmake.extend(['-Werror=dev', '-Werror=deprecated', '--warn-uninitialized', '..' if not buildType else '../..'])
    command = ' '.join(cmake)

    if not os.path.exists(basePath):
        os.makedirs(basePath)
    elif 'force' in arguments:
        paths = os.listdir(basePath)
        for path in paths:
            if path.lower().startswith('cmake'):
                full = basePath + '/' + path
                if os.path.isdir(full):
                    shutil.rmtree(full, ignore_errors=False)
                else:
                    os.remove(full)
        print('Cleared previous.')
    os.chdir(basePath)
    subprocess.call(command, shell=True)

    return 0



# PytgVoIP - Telegram VoIP Library for Python
# Copyright (C) 2019 bakatrouble <https://github.com/bakatrouble>
#
# This file is part of PytgVoIP.
#
# PytgVoIP is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PytgVoIP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with PytgVoIP.  If not, see <http://www.gnu.org/licenses/>.


from tgvoip.tgvoip import *
from tgvoip.tgvoip import __all__

__version__ = '0.0.5'



#!/usr/bin/env python3

# PytgVoIP - Telegram VoIP Library for Python
# Copyright (C) 2019 bakatrouble <https://github.com/bakatrouble>
#
# This file is part of PytgVoIP.
#
# PytgVoIP is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PytgVoIP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with PytgVoIP.  If not, see <http://www.gnu.org/licenses/>.


import multiprocessing
import os
import re
import sys
import platform
import subprocess

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext


def check_libraries():
    args = 'gcc -lssl -lopus'.split()
    with subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE) as process:
        stdout, stderr = process.communicate()
        match = re.findall(r'cannot find -l(\w+)', stderr.decode())
        if match:
            raise RuntimeError(
                'Following libraries are not installed: {}\nFor installation guide refer to '
                'https://pytgvoip.readthedocs.io/en/latest/guides/install.html'.format(', '.join(match))
            )


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError("CMake must be installed to build the following extensions: " +
                               ", ".join(e.name for e in self.extensions))

        if platform.system() != 'Windows':
            check_libraries()

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args = ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir,
                      '-DPYTHON_EXECUTABLE=' + sys.executable]

        cfg = 'Release'
        build_args = ['--config', cfg, '--target', '_tgvoip']

        if platform.system() == "Windows":
            cmake_args += ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}'.format(cfg.upper(), extdir)]
            if sys.maxsize > 2**32:
                cmake_args += ['-A', 'x64']
            else:
                cmake_args += ['-A', 'Win32']
            build_args += ['--', '/m:{}'.format(multiprocessing.cpu_count() + 1)]
        else:
            cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
            build_args += ['--', '-j{}'.format(multiprocessing.cpu_count() + 1)]

        env = os.environ.copy()
        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(env.get('CXXFLAGS', ''),
                                                              self.distribution.get_version())
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call(['cmake', ext.sourcedir] + cmake_args, cwd=self.build_temp, env=env)
        subprocess.check_call(['cmake', '--build', '.'] + build_args, cwd=self.build_temp)


def get_version():
    init_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'src', 'tgvoip', '__init__.py')
    with open(init_path, encoding='utf-8') as f:
        version = re.findall(r"__version__ = '(.+)'", f.read())[0]
        if os.environ.get('BUILD') is None and 'pip' not in __file__:
            version += '+develop'
        return version


def get_long_description():
    with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), 'README.md'), encoding='utf-8') as f:
        return f.read()


setup(
    name='pytgvoip',
    version=get_version(),
    license='LGPLv3+',
    author='bakatrouble',
    author_email='bakatrouble@gmail.com',
    description='Telegram VoIP Library for Python',
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    url='https://github.com/bakatrouble/pytgvoip',
    keywords='telegram messenger voip library python',
    project_urls={
        'Tracker': 'https://github.com/bakatrouble/pytgvoip/issues',
        'Community': 'https:/t.me/pytgvoip',
        'Source': 'https://github.com/bakatrouble/pytgvoip',
    },
    python_required='~=3.4',
    ext_modules=[CMakeExtension('_tgvoip')],
    packages=['tgvoip'],
    package_dir={'tgvoip': os.path.join('src', 'tgvoip')},
    package_data={'': [os.path.join('src', '_tgvoip.pyi')]},
    cmdclass={'build_ext': CMakeBuild},
    zip_safe=False,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: C++',
        'Topic :: Internet',
        'Topic :: Communications',
        'Topic :: Communications :: Internet Phone',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



# PytgVoIP - Telegram VoIP Library for Python
# Copyright (C) 2019 bakatrouble <https://github.com/bakatrouble>
#
# This file is part of PytgVoIP.
#
# PytgVoIP is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PytgVoIP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with PytgVoIP.  If not, see <http://www.gnu.org/licenses/>.

import json
import os
import sys
from datetime import datetime
from enum import Enum
from typing import Union, List

from _tgvoip import (
    NetType as _NetType,
    DataSaving as _DataSaving,
    CallState as _CallState,
    CallError as _CallError,
    Stats,
    Endpoint,
    VoIPController as _VoIPController,
    VoIPServerConfig as _VoIPServerConfig
)

from tgvoip.utils import get_real_elapsed_time


# docstring magic ahead


class NetType(Enum):
    """
    An enumeration of network types

    Members:
        * UNKNOWN = 0
        * GPRS = 1
        * EDGE = 2
        * NET_3G = 3
        * HSPA = 4
        * LTE = 5
        * WIFI = 6
        * ETHERNET = 7
        * OTHER_HIGH_SPEED = 8
        * OTHER_LOW_SPEED = 9
        * DIALUP = 10
        * OTHER_MOBILE = 11
    """

    UNKNOWN = _NetType.UNKNOWN
    GPRS = _NetType.GPRS
    EDGE = _NetType.EDGE
    NET_3G = _NetType.NET_3G
    HSPA = _NetType.HSPA
    LTE = _NetType.LTE
    WIFI = _NetType.WIFI
    ETHERNET = _NetType.ETHERNET
    OTHER_HIGH_SPEED = _NetType.OTHER_HIGH_SPEED
    OTHER_LOW_SPEED = _NetType.OTHER_LOW_SPEED
    DIALUP = _NetType.DIALUP
    OTHER_MOBILE = _NetType.OTHER_MOBILE


class CallState(Enum):
    """
    An enumeration of call states

    Members:
        * WAIT_INIT = 1
        * WAIT_INIT_ACK = 2
        * ESTABLISHED = 3
        * FAILED = 4
        * RECONNECTING = 5
        * HANGING_UP = 10
        * ENDED = 11
        * EXCHANGING_KEYS = 12
        * WAITING = 13
        * REQUESTING = 14
        * WAITING_INCOMING = 15
        * RINGING = 16
        * BUSY = 17
    """

    WAIT_INIT = _CallState.WAIT_INIT
    WAIT_INIT_ACK = _CallState.WAIT_INIT_ACK
    ESTABLISHED = _CallState.ESTABLISHED
    FAILED = _CallState.FAILED
    RECONNECTING = _CallState.RECONNECTING
    HANGING_UP = 10
    ENDED = 11
    EXCHANGING_KEYS = 12
    WAITING = 13
    REQUESTING = 14
    WAITING_INCOMING = 15
    RINGING = 16
    BUSY = 17


class DataSaving(Enum):
    """
    An enumeration of data saving modes

    Members:
        * NEVER = 0
        * MOBILE = 1
        * ALWAYS = 2
    """
    NEVER = _DataSaving.NEVER
    MOBILE = _DataSaving.MOBILE
    ALWAYS = _DataSaving.ALWAYS


class CallError(Enum):
    """
    An enumeration of call errors

    Members:
        * UNKNOWN = 0
        * INCOMPATIBLE = 1
        * TIMEOUT = 2
        * AUDIO_IO = 3
        * PROXY = 4
    """
    UNKNOWN = _CallError.UNKNOWN
    INCOMPATIBLE = _CallError.INCOMPATIBLE
    TIMEOUT = _CallError.TIMEOUT
    AUDIO_IO = _CallError.AUDIO_IO
    PROXY = _CallError.PROXY


class VoIPController(_VoIPController):
    """
    A wrapper around C++ wrapper for libtgvoip ``VoIPController``

    Args:
        persistent_state_file (``str``, *optional*): ?, empty to not use
        debug (``bool``, *optional*): Modifies logging behavior
        logs_dir (``str``, *optional*): Logs directory

    Class attributes:
        LIBTGVOIP_VERSION
            Used ``libtgvoip`` version

        CONNECTION_MAX_LAYER
            Maximum layer supported by used ``libtgvoip`` version

    Attributes:
        persistent_state_file:
            Value set in the constructor

        call_state_changed_handlers:
            ``list`` of call state change callbacks, callbacks receive a :class:`CallState` object as argument

        signal_bars_changed_handlers
            ``list`` of signal bars count change callbacks, callbacks receive an ``int`` object as argument
    """

    def __init__(self, persistent_state_file: str = '', debug=False, logs_dir='logs'):
        super().__init__(persistent_state_file)  # _VoIPController.__init__(self, persistent_state_file)
        self.debug = debug
        self.logs_dir = logs_dir
        self.start_time = 0
        self.send_audio_frame_callback = lambda length: b''
        self.recv_audio_frame_callback = lambda frame: ...
        self.call_state_changed_handlers = []
        self.signal_bars_changed_handlers = []
        self._init()

    @property
    def call_duration(self) -> int:
        """
        Current call duration in seconds as ``int`` if call was started, otherwise 0
        """
        return int(get_real_elapsed_time() - self.start_time) if self.start_time else 0

    def start(self):
        """
        Start the controller
        """
        super().start()

    def connect(self):
        """
        Start the call
        """
        super().connect()

    def set_proxy(self, address: str, port: int = 1080, username: str = '', password: str = ''):
        """
        Set SOCKS5 proxy config

        Args:
            address (``str``): Proxy hostname or IP address
            port (``int``, *optional*): Proxy port
            username (``int``, *optional*): Proxy username
            password (``int``, *optional*): Proxy password

        Raises:
            :class:`ValueError` if :attr:`address` is empty
        """
        if not address:
            raise ValueError('address can\'t be empty')
        super().set_proxy(address, port, username, password)

    def set_encryption_key(self, key: bytes, is_outgoing: bool):
        """
        Set call auth key

        Args:
            key (``bytes``): Auth key, must be exactly 256 bytes
            is_outgoing (``bool``): Is call outgoing

        Raises:
            :class:`ValueError` if provided auth key has wrong length
        """
        if len(key) != 256:
            raise ValueError('key length must be exactly 256 bytes but is {}'.format(len(key)))
        super().set_encryption_key(key, is_outgoing)

    def set_remote_endpoints(self, endpoints: List[Endpoint], allow_p2p: bool, tcp: bool, connection_max_layer: int):
        """
        Set remote endpoints received in call object from Telegram.

        Usually it's ``[call.connection] + call.alternative_connections``.

        You must build :class:`Endpoint` objects from MTProto :class:`phoneConnection` objects and pass them in list.

        Args:
            endpoints (``list`` of :class:`Endpoint`): List of endpoints
            allow_p2p (``bool``): Is p2p connection allowed, usually `call.p2p_allowed` value is used
            tcp (``bool``): Connect via TCP, not recommended
            connection_max_layer (``int``): Use a value provided by :attr:`VoIPController.CONNECTION_MAX_LAYER`

        Raises:
            :class:`ValueError` if either no endpoints are provided or endpoints without IPv4 or with wrong \
            ``peer_tag`` (must be either ``None`` or have length of 16 bytes) are detected
        """
        if not endpoints:
            raise ValueError('endpoints len is 0')
        for ep in endpoints:
            if ep.ip is None or not len(ep.ip):
                raise ValueError('endpoint {} has empty/null ipv4'.format(ep))
            if ep.peer_tag is not None and len(ep.peer_tag) != 16:
                raise ValueError('endpoint {} has peer_tag of wrong length'.format(ep))
        super().set_remote_endpoints(endpoints, allow_p2p, tcp, connection_max_layer)

    def get_debug_string(self) -> str:
        """
        Get debug string

            Returns:
                ``str`` containing debug info
        """
        return super().get_debug_string()

    def set_network_type(self, _type: NetType):
        """
        Set network type

        Args:
            _type (:class:`NetType`): Network type to set
        """
        super().set_network_type(_NetType(_type.value))

    def set_mic_mute(self, mute: bool):
        """
        Set "microphone" state. If muted, audio is not being sent

        Args:
            mute (``bool``): Whether to mute "microphone"
        """
        super().set_mic_mute(mute)

    def set_config(self,
                   recv_timeout: float,
                   init_timeout: float,
                   data_saving_mode: DataSaving,
                   call_id: int,
                   enable_aec: bool = True,
                   enable_ns: bool = True,
                   enable_agc: bool = True,
                   log_file_path: str = None,
                   status_dump_path: str = None,
                   log_packet_stats: bool = None):
        """
        Set call config

        Args:
            recv_timeout (``float``):
                Packet receive timeout, usually value received from ``help.getConfig()`` is used

            init_timeout (``float``):
                Packet init timeout, usually value received from ``help.getConfig()`` is used

            data_saving_mode (:class:`DataSaving`): Data saving mode

            call_id (``int``): Call ID

            enable_aec (``bool``, *optional*):
                Whether to enable automatic echo cancellation, defaults to ``True``

            enable_ns (``bool``, *optional*):
                Whether to enable noise suppression, defaults to ``True``

            enable_agc (``bool``, *optional*):
                Whether to enable automatic gain control, defaults to ``True``

            log_file_path (``str``, *optional*):
                Call log file path, calculated automatically if not provided

            status_dump_path (``str``, *optional*):
                Status dump path, calculated automatically if not provided and ``debug`` is enabled

            log_packet_stats (``bool``, *optional*):
                Whether to log packet stats, defaults to ``debug`` value
        """
        if log_file_path is None:
            if self.debug:
                log_file_path = self._get_log_file_path('voip{}'.format(call_id))
            else:
                log_file_path = self._get_log_file_path_for_call_id(call_id)  # wtf?
        if status_dump_path is None:
            status_dump_path = self._get_log_file_path('voip_stats') if self.debug else ''
        if log_packet_stats is None:
            log_packet_stats = self.debug
        super().set_config(recv_timeout, init_timeout, _DataSaving(data_saving_mode.value), enable_aec, enable_ns, enable_agc,
                           log_file_path, status_dump_path, log_packet_stats)

    def debug_ctl(self, request: int, param: int):
        """
        Debugging options

        Args:
            request (``int``):
                Option (``1`` for max bitrate, ``2`` for packet loss (in percents), ``3`` for toggling p2p, \
                ``4`` for toggling echo cancelling)

            param (``int``):
                Numeric value for options 1 and 2, ``0`` or ``1`` for options 3 and 4
        """
        super().debug_ctl(request, param)

    def get_preferred_relay_id(self) -> int:
        """
        Get preferred relay ID (used in ``discardCall`` MTProto request)

        Returns:
            ``int`` ID
        """
        return super().get_preferred_relay_id()

    def get_last_error(self) -> CallError:
        """
        Get last error type

        Returns:
            :class:`CallError` matching last occurred error type
        """
        return CallError(super().get_last_error())

    def get_stats(self) -> Stats:
        """
        Get call stats

        Returns:
            :class:`Stats` object
        """
        return super().get_stats()

    def get_debug_log(self) -> str:
        """
        Get debug log

        Returns:
            JSON ``str`` containing debug log
        """
        return super().get_debug_log()

    def set_audio_output_gain_control_enabled(self, enabled: bool):
        """
        Toggle output gain control

        Args:
            enabled (``bool``): Whether to enable output gain control
        """
        super().set_audio_output_gain_control_enabled(enabled)

    def set_echo_cancellation_strength(self, strength: int):
        """
        Set echo cancellation strength, does nothing currently but was in Java bindings (?)

        Args:
            strength (``int``): Strength value
        """
        super().set_echo_cancellation_strength(strength)

    def get_peer_capabilities(self) -> int:
        """
        Get peer capabilities

        Returns:
            ``int`` with bit mask, looks like it is used only for experimental features (group, video calls)
        """
        return super().get_peer_capabilities()

    def need_rate(self) -> bool:
        """
        Get whether the call needs rating

        Returns:
            ``bool`` value
        """
        return super().need_rate()

    # native code callback
    def _handle_state_change(self, state: _CallState):
        state = CallState(state)

        if state == CallState.ESTABLISHED and not self.start_time:
            self.start_time = get_real_elapsed_time()

        for handler in self.call_state_changed_handlers:
            callable(handler) and handler(state)

    # native code callback
    def _handle_signal_bars_change(self, count: int):
        for handler in self.signal_bars_changed_handlers:
            callable(handler) and handler(count)

    def update_state(self, state: CallState):
        """
        Manually update state (only triggers handlers)

        Args:
            state (:class:`CallState`): State to set
        """
        self._handle_state_change(state)

    def set_send_audio_frame_callback(self, func: callable):
        """
        Set callback providing audio data to send

        Should accept one argument (``int`` length of requested audio frame) and return ``bytes`` object with audio \
        data encoded in 16-bit signed PCM

        If returned object has insufficient length, it will be automatically padded with zero bytes

        Args:
            func (``callable``): Callback function
        """
        self.send_audio_frame_callback = func

    def _send_audio_frame_impl(self, length: int):
        frame = b''
        if callable(self.send_audio_frame_callback):
            frame = self.send_audio_frame_callback(length)
        return frame.ljust(length, b'\0')

    def set_recv_audio_frame_callback(self, func: callable):
        """
        Set callback receiving incoming audio data

        Should accept one argument (``bytes``) with audio data encoded in 16-bit signed PCM

        Args:
            func (``callable``): Callback function
        """
        self.recv_audio_frame_callback = func

    def _recv_audio_frame_impl(self, frame: bytes):
        if callable(self.recv_audio_frame_callback):
            self.recv_audio_frame_callback(frame)

    def _get_log_file_path(self, name: str) -> str:
        os.makedirs(self.logs_dir, exist_ok=True)
        now = datetime.now()
        fname = '{}_{}_{}_{}_{}_{}_{}.txt'.format(now.year, now.month, now.day, now.hour, now.minute, now.second, name)
        return os.path.abspath(os.path.join(self.logs_dir, fname))

    def _get_log_file_path_for_call_id(self, call_id: int) -> str:
        os.makedirs(self.logs_dir, exist_ok=True)
        # Java version cleans up old logs (*.log) for non-debug version here (leaves 20 latest)
        return os.path.abspath(os.path.join(self.logs_dir, '{}.log'.format(call_id)))


class VoIPServerConfig(_VoIPServerConfig):
    """
    Global server config class. This class contains default config in its source
    """

    # default config
    config = {
        'audio_max_bitrate': 20000,
        'audio_max_bitrate_gprs': 8000,
        'audio_max_bitrate_edge': 16000,
        'audio_max_bitrate_saving': 8000,
        'audio_init_bitrate': 16000,
        'audio_init_bitrate_gprs': 8000,
        'audio_init_bitrate_edge': 8000,
        'audio_init_bitrate_saving': 8000,
        'audio_bitrate_step_incr': 1000,
        'audio_bitrate_step_decr': 1000,
        'audio_min_bitrate': 8000,
        'relay_switch_threshold': 0.8,
        'p2p_to_relay_switch_threshold': 0.6,
        'relay_to_p2p_switch_threshold': 0.8,
        'reconnecting_state_timeout': 2.0,
        'rate_flags': 0xFFFFFFFF,
        'rate_min_rtt': 0.6,
        'rate_min_send_loss': 0.2,
        'packet_loss_for_extra_ec': 0.02,
        'max_unsent_stream_packets': 2,
    }

    @classmethod
    def set_config(cls, _json: Union[str, dict]):
        """
        Set global server config

        Args:
            _json (``str`` | ``dict``): either JSON-encoded object or ``dict`` containing config values. \
            Might be received from MTProto ``phone.getCallConfig()`` call, if not set default values are used

        Raises:
            Prints an error to ``stderr`` if JSON parsing (for ``str`` argument) or encoding (for ``dict`` argument) \
            has occurred
        """
        try:
            if isinstance(_json, dict):
                _json = json.dumps(_json)
            cls.config.update(json.loads(_json))
            _VoIPServerConfig.set_config(_json)
        except json.JSONDecodeError as e:
            print('Error parsing VoIP config', e, file=sys.stderr)
        except TypeError as e:
            print('Error building JSON', e, file=sys.stderr)

    @classmethod
    def set_bitrate_config(cls, init_bitrate: int = 16000, max_bitrate: int = 20000, min_bitrate: int = 8000,
                           decrease_step: int = 1000, increase_step: int = 1000):
        """
        Helper method for setting bitrate options

        Args:
            init_bitrate (``int``): Initial bitrate value
            max_bitrate (``int``): Maximum bitrate value
            min_bitrate (``int``):  Minimum bitrate value
            decrease_step (``int``): Bitrate decrease step
            increase_step (``int``): Bitrate increase step

        Raises:
            Same as :meth:`set_config`
        """
        cls.set_config({
            'audio_init_bitrate': init_bitrate,
            'audio_max_bitrate': max_bitrate,
            'audio_min_bitrate': min_bitrate,
            'audio_bitrate_step_decr': decrease_step,
            'audio_bitrate_step_incr': increase_step,
        })


__all__ = ['NetType', 'DataSaving', 'CallState', 'CallError', 'Stats', 'Endpoint', 'VoIPController',
           'VoIPServerConfig']
